# -*- coding: utf-8 -*-
"""
@file
@brief Implements function @see fn run_cmd.

.. versionadded:: 1.1
"""
from __future__ import print_function
import sys
import os
import time
import subprocess
import threading
import warnings
import shlex
if sys.version_info[0] == 2:
    import Queue as queue
else:
    import queue


class RunCmdException(Exception):
    """
    raised by function @see fn run_cmd
    """
    pass


def get_interpreter_path():
    """
    return the interpreter path
    """
    if sys.platform.startswith("win"):
        return sys.executable.replace("pythonw.exe", "python.exe")
    else:
        return sys.executable


def split_cmp_command(cmd, remove_quotes=True):
    """
    splits a command line
    @param      cmd             command line
    @param      remove_quotes   True by default
    @return                     list
    """
    if isinstance(cmd, str  # unicode#
                  ):
        return shlex.split(cmd)
    else:
        return cmd


def decode_outerr(outerr, encoding, encerror, msg):
    """
    decode the output or the error after running a command line instructions

    @param      outerr      output or error
    @param      encoding    encoding (if None, it is replaced by ascii)
    @param      encerror    how to handle errors
    @param      msg         to add to the exception message
    @return                 converted string

    .. versionchanged:: 1.4
        If *encoding* is None, it is replaced by ``'ascii'``.
    """
    if encoding is None:
        encoding = "ascii"
    typstr = str  # unicode#
    if not isinstance(outerr, bytes):
        raise TypeError(
            "only able to decode bytes, not " + typstr(type(outerr)))
    try:
        out = outerr.decode(encoding, errors=encerror)
        return out
    except UnicodeDecodeError as exu:
        try:
            out = outerr.decode(
                "utf8" if encoding != "utf8" else "latin-1", errors=encerror)
            return out
        except Exception as e:
            out = outerr.decode(encoding, errors='ignore')
            raise Exception("issue with cmd (" + encoding + "):" +
                            typstr(msg) + "\n" + typstr(exu) + "\n-----\n" + out) from e
    raise Exception("complete issue with cmd:" + typstr(msg))


def skip_run_cmd(cmd, sin="", shell=True, wait=False, log_error=True,
                 stop_running_if=None, encerror="ignore",
                 encoding="utf8", change_path=None, communicate=True,
                 preprocess=True, timeout=None, catch_exit=False, fLOG=None,
                 timeout_listen=None, tell_if_no_output=None):
    """
    has the same signature as @see fn run_cmd but does nothing

    .. versionadded:: 1.0
    """
    return "", ""


def run_cmd_private(cmd, sin="", shell=True, wait=False, log_error=True,
                    stop_running_if=None, encerror="ignore", encoding="utf8",
                    change_path=None, communicate=True, preprocess=True, timeout=None,
                    catch_exit=False, fLOG=None, tell_if_no_output=None):
    """
    run a command line and wait for the result
    @param      cmd                 command line
    @param      sin                 sin: what must be written on the standard input
    @param      shell               if True, cmd is a shell command (and no command window is opened)
    @param      wait                call ``proc.wait``
    @param      log_error           if log_error, call fLOG (error)
    @param      stop_running_if     the function stops waiting if some condition is fulfilled.
                                    The function received the last line from the logs.
                                    Signature: ``stop_waiting_if(last_out, last_err) -> bool``.
                                    The function must return True to stop waiting.
                                    This function can also be used to intercept the standard output
                                    and the standard error while running.
    @param      encerror            encoding errors (ignore by default) while converting the output into a string
    @param      encoding            encoding of the output
    @param      change_path         change the current path if not None (put it back after the execution)
    @param      communicate         use method `communicate <https://docs.python.org/3/library/subprocess.html#subprocess.Popen.communicate>`_
                                    which is supposed to be safer,
                                    parameter ``wait`` must be True
    @param      preprocess          preprocess the command line if necessary (not available on Windows) (False to disable that option)
    @param      timeout             when data is sent to stdin (``sin``), a timeout is needed to avoid
                                    waiting for ever (*timeout* is in seconds)
    @param      catch_exit          catch *SystemExit* exception
    @param      fLOG                logging function (if not None, bypass others parameters)
    @param      tell_if_no_output   tells if there is no output every *tell_if_no_output* seconds
    @return                         content of stdout, stdres  (only if wait is True)

    .. exref::
        :title: Run a program using the command line

        ::

            from pyquickhelper.loghelper import run_cmd
            out,err = run_cmd( "python setup.py install", wait=True)

    If you are using this function to run git function, parameter ``shell`` must be True.

    .. versionadded:: 1.1
    """
    if fLOG is not None:
        fLOG("execute", cmd)

    if sys.platform.startswith("win"):
        cmdl = cmd
    else:
        cmdl = split_cmp_command(cmd) if preprocess else cmd

    if catch_exit:
        try:
            pproc = subprocess.Popen(cmdl,
                                     shell=shell,
                                     stdin=subprocess.PIPE if (
                                         sin and len(sin) > 0) else None,
                                     stdout=subprocess.PIPE if wait else None,
                                     stderr=subprocess.PIPE if wait else None,
                                     cwd=change_path)
        except SystemExit as e:
            raise RunCmdException("SystemExit raised (1)") from e

    else:
        pproc = subprocess.Popen(cmdl,
                                 shell=shell,
                                 stdin=subprocess.PIPE if (
                                     sin and len(sin) > 0) else None,
                                 stdout=subprocess.PIPE if wait else None,
                                 stderr=subprocess.PIPE if wait else None,
                                 cwd=change_path)

    if isinstance(cmd, list):
        cmd = " ".join(cmd)

    if wait:
        skip_out_err = False
        out = []
        err = []
        err_read = False
        skip_waiting = False

        if communicate:
            # communicate is True
            if tell_if_no_output is not None:
                raise NotImplementedError(
                    "tell_if_no_output is not implemented when communicate is True")
            if stop_running_if is not None:
                raise NotImplementedError(
                    "stop_running_if is not implemented when communicate is True")
            input = None if (sin is None or len(sin) > 0) else sin.encode()
            if input is not None and len(input) > 0:
                if fLOG is not None:
                    fLOG("input", [input])

            if catch_exit:
                try:
                    if sys.version_info[0] == 2:
                        if timeout is not None:
                            raise NotImplementedError(
                                "timeout is only available with Python 3")
                        stdoutdata, stderrdata = pproc.communicate(input=input)
                    else:
                        stdoutdata, stderrdata = pproc.communicate(
                            input=input, timeout=timeout)
                except SystemExit as e:
                    raise RunCmdException("SystemExit raised (2)") from e
            else:
                if sys.version_info[0] == 2:
                    if timeout is not None:
                        warnings.warn(
                            "[run_cmd_private] timeout is only available with Python 3")
                    stdoutdata, stderrdata = pproc.communicate(input=input)
                else:
                    stdoutdata, stderrdata = pproc.communicate(
                        input=input, timeout=timeout)

            out = decode_outerr(stdoutdata, encoding, encerror, cmd)
            err = decode_outerr(stderrdata, encoding, encerror, cmd)
        else:
            # communicate is False: use of threads
            if sin is not None and len(sin) > 0:
                raise Exception(
                    "communicate should be True to send something on stdin")
            stdout, stderr = pproc.stdout, pproc.stderr

            begin = time.perf_counter()
            last_update = begin
            # with threads
            (stdoutReader, stdoutQueue) = _AsyncLineReader.getForFd(
                stdout, catch_exit=catch_exit)
            (stderrReader, stderrQueue) = _AsyncLineReader.getForFd(
                stderr, catch_exit=catch_exit)
            runloop = True

            while (not stdoutReader.eof() or not stderrReader.eof()) and runloop:
                while not stdoutQueue.empty():
                    line = stdoutQueue.get()
                    decol = decode_outerr(
                        line, encoding, encerror, cmd)
                    if fLOG is not None:
                        fLOG(decol.strip("\n\r"))
                    out.append(decol.strip("\n\r"))
                    last_update = time.perf_counter()
                    if stop_running_if is not None and stop_running_if(decol, None):
                        runloop = False
                        break

                while not stderrQueue.empty():
                    line = stderrQueue.get()
                    decol = decode_outerr(
                        line, encoding, encerror, cmd)
                    if fLOG is not None:
                        fLOG(decol.strip("\n\r"))
                    err.append(decol.strip("\n\r"))
                    last_update = time.perf_counter()
                    if stop_running_if is not None and stop_running_if(None, decol):
                        runloop = False
                        break
                time.sleep(0.05)

                delta = time.perf_counter() - last_update
                if tell_if_no_output is not None and delta >= tell_if_no_output:
                    if fLOG is not None:
                        fLOG("[run_cmd] No update in {0} seconds for cmd: {1}".format(
                            "%5.1f" % (last_update - begin), cmd))
                    last_update = time.perf_counter()
                full_delta = time.perf_counter() - begin
                if timeout is not None and full_delta > timeout:
                    runloop = False
                    if fLOG is not None:
                        fLOG("[run_cmd] Timeout after {0} seconds for cmd: {1}".format(
                            "%5.1f" % full_delta, cmd))
                    break

            if runloop:
                # Waiting for async readers to finish...
                stdoutReader.join()
                stderrReader.join()

                # Waiting for process to exit...
                returnCode = pproc.wait()
                err_read = True

                if returnCode != 0:
                    try:
                        # we try to close the ressources
                        stdout.close()
                        stderr.close()
                    except Exception as e:
                        warnings.warn("Unable to close stdout and sterr.")
                    if catch_exit:
                        raise RunCmdException("SystemExit raised with error code {0}\nOUT:\n{1}\nERR-0:\n{2}".format(
                            returnCode, "\n".join(out), "\n".join(err)))
                    raise subprocess.CalledProcessError(returnCode, cmd)

                if not skip_waiting:
                    pproc.wait()
            else:
                out.append("[run_cmd] killing process.")
                if fLOG is not None:
                    fLOG(
                        "[run_cmd] killing process because stop_running_if returned True.")
                pproc.kill()
                err_read = True
                if fLOG is not None:
                    fLOG("[run_cmd] process killed.")
                skip_out_err = True

            out = "\n".join(out)
            if skip_out_err:
                err = "Process killed."
            else:
                if err_read:
                    err = "\n".join(err)
                else:
                    temp = err = stderr.read()
                    try:
                        err = decode_outerr(temp, encoding, encerror, cmd)
                    except Exception:
                        err = decode_outerr(temp, encoding, "ignore", cmd)
                stdout.close()
                stderr.close()

        # same path for whether communicate is False or True
        err = err.replace("\r\n", "\n")
        if fLOG is not None:
            fLOG("end of execution", cmd)

        if len(err) > 0 and log_error and fLOG is not None:
            fLOG("error (log)\n%s" % err)

        if sys.platform.startswith("win"):
            return out.replace("\r\n", "\n"), err.replace("\r\n", "\n")
        else:
            return out, err
    else:

        return "", ""


class _AsyncLineReader(threading.Thread):

    def __init__(self, fd, outputQueue, catch_exit):
        threading.Thread.__init__(self)

        assert isinstance(outputQueue, queue.Queue)
        assert callable(fd.readline)

        self.fd = fd
        self.catch_exit = catch_exit
        self.outputQueue = outputQueue

    def run(self):
        if self.catch_exit:
            try:
                for _ in map(self.outputQueue.put, iter(self.fd.readline, b'')):
                    pass
            except SystemExit as e:
                self.outputQueue.put(str(e))
                raise RunCmdException("SystemExit raised (3)") from e
        else:
            for _ in map(self.outputQueue.put, iter(self.fd.readline, b'')):
                pass

    def eof(self):
        return not self.is_alive() and self.outputQueue.empty()

    @classmethod
    def getForFd(cls, fd, start=True, catch_exit=False):
        q = queue.Queue()
        reader = cls(fd, q, catch_exit)

        if start:
            reader.start()

        return reader, q


def run_cmd_old(cmd, sin="", shell=False, wait=False, log_error=True,
                secure=None, stop_waiting_if=None, do_not_log=False,
                encerror="ignore", encoding="utf8", cwd=None, fLOG=print):
    """
    run a command line and wait for the result
    @param      cmd                 command line
    @param      sin                 sin, what must be written on the standard input
    @param      shell               if True, cmd is a shell command (and no command window is opened)
    @param      wait                call proc.wait
    @param      log_error           if log_error, call fLOG (error)
    @param      secure              if secure is a string (a valid filename), the function stores the output in a file
                                    and reads it continuously
    @param      stop_waiting_if     the function stops waiting if some condition is fulfilled.
                                    The function received the last line from the logs.
    @param      do_not_log          do not log the output
    @param      encerror            encoding errors (ignore by default) while converting the output into a string
    @param      encoding            encoding of the output
    @param      cwd                 current folder
    @param      fLOG                logging function
    @return                         content of stdout, stderr  (only if wait is True)


    .. faqref::
        :title: Exception when installing a module

        This error can occur when a module is installed on a virtual environment
        created before *pip* was updated on the main distribution.
        The solution consists in removing the virtual environment and create it again.

        ::

            c:\\Python34_x64vir\\install\\Scripts\\python -u setup.py install
            running install
            running bdist_egg
            running egg_info
            Traceback (most recent call last):
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2785, in _dep_map
                return self.__dep_map
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2642, in __getattr__
                raise AttributeError(attr)
            AttributeError: _DistInfoDistribution__dep_map

            During handling of the above exception, another exception occurred:

            Traceback (most recent call last):
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2776, in _parsed_pkg_info
                return self._pkg_info
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2642, in __getattr__
                raise AttributeError(attr)
            AttributeError: _pkg_info

            During handling of the above exception, another exception occurred:

            Traceback (most recent call last):
            File "setup.py", line 169, in <module>
                package_data=package_data,
            File "C:\\Python34_x64\\Lib\\distutils\\core.py", line 148, in setup
                dist.run_commands()
            File "C:\\Python34_x64\\Lib\\distutils\\dist.py", line 955, in run_commands
                self.run_command(cmd)
            File "C:\\Python34_x64\\Lib\\distutils\\dist.py", line 974, in run_command
                cmd_obj.run()
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\setuptools\\command\\install.py", line 67, in run
                self.do_egg_install()
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\setuptools\\command\\install.py", line 109, in do_egg_install
                self.run_command('bdist_egg')
            File "C:\\Python34_x64\\Lib\\distutils\\cmd.py", line 313, in run_command
                self.distribution.run_command(command)
            File "C:\\Python34_x64\\Lib\\distutils\\dist.py", line 974, in run_command
                cmd_obj.run()
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\setuptools\\command\\bdist_egg.py", line 151, in run
                self.run_command("egg_info")
            File "C:\\Python34_x64\\Lib\\distutils\\cmd.py", line 313, in run_command
                self.distribution.run_command(command)
            File "C:\\Python34_x64\\Lib\\distutils\\dist.py", line 974, in run_command
                cmd_obj.run()
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\setuptools\\command\\egg_info.py", line 171, in run
                ep.require(installer=installer)
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2355, in require
                items = working_set.resolve(reqs, env, installer)
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 835, in resolve
                new_requirements = dist.requires(req.extras)[::-1]
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2586, in requires
                dm = self._dep_map
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2787, in _dep_map
                self.__dep_map = self._compute_dependencies()
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2809, in _compute_dependencies
                for req in self._parsed_pkg_info.get_all('Requires-Dist') or []:
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 2778, in _parsed_pkg_info
                metadata = self.get_metadata(self.PKG_INFO)
            File "c:\\Python34_x64vir\\install\\lib\\site-packages\\pkg_resources\\__init__.py", line 1993, in get_metadata
                raise KeyError("No metadata except PKG-INFO is available")
            KeyError: 'No metadata except PKG-INFO is available'
    """
    if sin is not None and sin != "":
        raise NotImplementedError("sin is not used")

    if secure is not None:
        if not do_not_log:
            fLOG("secure=", secure)
        with open(secure, "w") as f:
            f.write("")
        add = ">%s" % secure
        if isinstance(cmd, str):
            cmd += " " + add
        else:
            cmd.append(add)
    if not do_not_log and fLOG is not None:
        fLOG("execute ", cmd)

    if sys.platform.startswith("win"):

        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        try:
            proc = subprocess.Popen(cmd, shell=shell,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                    startupinfo=startupinfo, cwd=cwd)
        except FileNotFoundError as e:
            raise RunCmdException("unable to run CMD:\n{0}".format(cmd)) from e
    else:
        try:
            proc = subprocess.Popen(split_cmp_command(cmd), shell=shell,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                    cwd=cwd)
        except FileNotFoundError as e:
            raise RunCmdException("unable to run CMD:\n{0}".format(cmd)) from e
    if wait:

        out = []
        skip_waiting = False

        if secure is None:
            for line in proc.stdout:
                if not do_not_log and fLOG is not None:
                    fLOG(line.decode(encoding, errors=encerror).strip("\n"))
                try:
                    out.append(
                        line.decode(
                            encoding,
                            errors=encerror).strip("\n"))
                except UnicodeDecodeError as exu:
                    raise RunCmdException(
                        "issue with cmd:" +
                        str(cmd) +
                        "\n" +
                        str(exu))
                if proc.stdout.closed:
                    break
                if stop_waiting_if is not None and stop_waiting_if(
                        line.decode("utf8", errors=encerror)):
                    skip_waiting = True
                    break
        else:
            last = []
            while proc.poll() is None:
                if os.path.exists(secure):
                    with open(secure, "r") as f:
                        lines = f.readlines()
                    if len(lines) > len(last):
                        for line in lines[len(last):]:
                            if not do_not_log and fLOG is not None:
                                fLOG(line.strip("\n"))
                            out.append(line.strip("\n"))
                        last = lines
                    if stop_waiting_if is not None and len(
                            last) > 0 and stop_waiting_if(last[-1]):
                        skip_waiting = True
                        break
                time.sleep(0.1)

        if not skip_waiting:
            proc.wait()

        out = "\n".join(out)
        err = proc.stderr.read().decode(encoding, errors=encerror)
        if not do_not_log and fLOG is not None:
            fLOG("end of execution ", cmd)
        if len(err) > 0 and log_error and not do_not_log and fLOG is not None:
            fLOG("error (log)\n%s" % err)
        # return bytes.decode (out, errors="ignore"), bytes.decode(err,
        # errors="ignore")
        proc.stdout.close()
        proc.stderr.close()

        return out, err
    else:
        return "", ""



"""
@brief      test log(time=20s)
"""
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import is_travis_or_appveyor
from pymyinstall.installhelper import get_installed_modules


class TestStatusHelper(unittest.TestCase):

    def test_status_helper(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        res = get_installed_modules(
            fLOG=fLOG, stop=10, pypi=True, short_list=["autopep8"])
        for f in res:
            fLOG(f)
        if is_travis_or_appveyor():
            self.assertEqual(len(res), 1)
        else:
            warnings.warn(
                "This test does not work well on a virtual environment [TestStatusHelper].")

        res = get_installed_modules(
            fLOG=fLOG, stop=5, pypi=True, short_list=["dataspyre"])
        fLOG(res)

        res = get_installed_modules(fLOG=fLOG, stop=5, pypi=True)
        for f in res:
            fLOG(f)
        if len(res) != 5:
            raise Exception("\n".join(
                "\n---------\n{0}/{1} {2}".format(i + 1, len(res), _) for i, _ in enumerate(res)))


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief Shortuts
"""
from .install_custom_exceptions import ManualDownloadException
from .install_custom import download_page, where_in_path
from .install_custom_7z import install_7z
from .install_custom_chromedriver import install_chromedriver
from .install_custom_git import install_git
from .install_custom_graphviz import install_graphviz
from .install_custom_javajdk import install_javajdk
from .install_custom_jenkins import install_jenkins
from .install_custom_julia import install_julia
from .install_custom_inkscape import install_inkscape
from .install_custom_miktex import install_miktex
from .install_custom_mingw import install_mingw
from .install_custom_operadriver import install_operadriver
from .install_custom_pandoc import install_pandoc
from .install_custom_putty import install_putty
from .install_custom_python import install_python, folder_older_than
from .install_custom_R import install_R
from .install_custom_scite import install_scite, modify_scite_properties
from .install_custom_sqlitespy import install_sqlitespy
from .install_custom_sbt import install_scala_sbt
from .install_custom_tdm_gcc import install_tdm_gcc



"""
@file
@brief Patch the distribution to make it portable
"""
from __future__ import print_function

import os
import re


class ShebangException(Exception):
    """
    exception raised when the shebang is not correct
    """
    pass


def win_patch_paths(folder, path_to_python="", fLOG=print):
    """
    Paths are absolute when they are installed in all scripts *.exe*,
    we replaced them either by an empty string (``path_to_python == ""``)
    or the current folder.

    @param      folder          folder where to find the executable
    @param      path_to_python  new python path (replace by)
    @param      fLOG            logging function
    @return                     list of tuple ('exe or py', 'modified file')

    See `Create a portable Python with Pip on Windows <http://www.clemens-sielaff.com/create-a-portable-python-with-pip-on-windows/>`_
    The first three parameters can be environment variables.
    They will be replaced by their values.
    """
    if folder in os.environ:
        folder = os.environ[folder]
    if path_to_python not in (None, "") and path_to_python in os.environ:
        path_to_python = os.environ[path_to_python]
    if path_to_python is None:
        exe = os.path.join(folder, "python.exe")
        if os.path.exists(exe):
            path_to_python = os.path.absapth(folder)
        else:
            exe = os.path.join(folder, "..", "python.exe")
            path_to_python = os.path.normpath(
                os.path.abspath(os.path.join(folder, "..")))

    files = os.listdir(folder)

    if len(path_to_python) > 0 and not path_to_python.endswith("\\"):
        path_to_python += "\\"

    pattern = "([#][!]((([A-Za-z][:])?[/\\\\]?[-a-zA-Z0-9_.]+[/\\\\])*?)(pythonw?[.]exe))"
    reg_exe = re.compile(pattern, re.IGNORECASE)
    breg_exe = re.compile(bytes(pattern, encoding="ascii"), re.IGNORECASE)
    g_path_to_python = path_to_python.replace("\\", "\\\\")
    bg_path_to_python = bytes(g_path_to_python, encoding="ascii")
    b1 = bytes("#!", encoding="ascii")
    b2 = bytes("\\5", encoding="ascii")

    operations = []
    for file in files:
        full = os.path.join(folder, file)
        if os.path.isfile(full):
            ext = os.path.splitext(full)[-1]

            if ext in {".py", ""}:
                with open(full, "r") as f:
                    content = f.read()
                new_content = reg_exe.sub(
                    "#!" + g_path_to_python + "\\5", content)
                if new_content != content:
                    fLOG("[pymy] update ", full)
                    operations.append(("update", full))
                    with open(full, "w") as f:
                        f.write(new_content)

            elif ext == ".exe":
                with open(full, "rb") as f:
                    content = f.read()
                new_content = breg_exe.sub(
                    b1 + bg_path_to_python + b2, content)
                if new_content != content:
                    fLOG("[win_patch_paths] update ", full)
                    operations.append(("update", full))
                    try:
                        with open(full, "wb") as f:
                            f.write(new_content)
                    except PermissionError as e:
                        with open(full + ".replace", "wb") as f:
                            f.write(new_content)
                        mes = ("unable to overwrite '{0}', it will have to be manually done, " +
                               "another file was created with .replace as an extension").format(
                            full)
                        raise Exception(mes) from e
            else:
                pass

    return operations



"""
@file
@brief Various functions to install `SciTE <http://www.scintilla.org/SciTE.html>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import unzip_files
from .install_custom import download_page, download_from_sourceforge, download_file
from ..installhelper.link_shortcuts import add_shortcut_to_desktop, suffix

if sys.version_info[0] == 2:
    from codecs import open


def IsSciteInstalled(dest_folder):
    """
    check if Scite was already installed

    @param      dest_folder     where it was installed
    @return                     boolean
    """
    if sys.platform.startswith("win"):
        file = os.path.join(dest_folder, "wscite", "SciTE.exe")
        return os.path.exists(file)
    else:
        raise NotImplementedError("not available on platform " + sys.platform)


def install_scite(dest_folder=".", fLOG=print, install=True, change_python_path=False, version=None):
    """
    install `SciTE <http://www.scintilla.org/SciTE.html>`_ (only on Windows)

    @param      dest_folder         where to download the setup
    @param      fLOG                logging function
    @param      install             install (otherwise only download)
    @param      change_python_path  update python path as well (if *install* is True)
    @param      version             version to install (unused)
    @return                         temporary file

    .. exref::
        :title: install SciTE

        The function downloads the latest version of SciTE.
        It also changes some settings for Python (no tabs, Courier New as a police).

        ::

            install_scite("my_folder_for_scite")

    .. versionchanged:: 1.1
        Parameter *change_python_path* was added.
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    if IsSciteInstalled(dest_folder):
        return os.path.join(
            os.path.abspath(dest_folder), "wscite", "SciTE.exe")

    if not sys.platform.startswith("win"):
        raise NotImplementedError(
            "SciTE can only be installed on Windows at the moment")

    url = "http://www.scintilla.org/SciTEDownload.html"
    page = download_page(url)

    rel = re.compile("Release ([0-9.]+)")
    rel = rel.findall(page)
    if len(rel) == 0:
        raise Exception("unable to find the release version")
    rel = rel[0]
    fLOG("[pymy] SciTE, release version ", rel)

    reg = re.compile("<a href=\\\"(.*zip.*)\\\">full download</a>")
    find = reg.findall(page)
    if len(find) != 1:
        raise Exception("unable to find the file to download at " +
                        url + "\nfound: " + str(len(find)) + "\n" + "\n".join(find))

    # should be something like http://www.scintilla.org/wscite356.zip
    newurl = find[0]
    outfile = os.path.join(dest_folder, "scite.zip")
    if not os.path.exists(outfile):
        download_file(newurl, outfile)

        if os.path.exists(outfile):
            file = outfile
        else:
            newurl = "http://sourceforge.net/projects/scintilla/files/SciTE/{0}/wscite{1}.zip/download?use_mirror=autoselect".format(
                rel,
                rel.replace(
                    ".",
                    ""))
            fLOG("[pymy] SciTE, download from ", newurl)
            file = download_from_sourceforge(
                newurl,
                outfile,
                fLOG=fLOG,
                temp_folder=dest_folder)

    if install:
        unzip_files(file, whereTo=dest_folder, fLOG=fLOG)
        modify_scite_properties(sys.executable if change_python_path else None,
                                os.path.join(dest_folder, "wscite"))
        return os.path.join(os.path.abspath(dest_folder), "wscite", "SciTE.exe")
    else:
        return outfile


def modify_scite_properties(python_path, scite_path):
    """
    modifies the scite properties

    @param      python_path     python path
    @param      scite_path      scrite path

    Avoid tabulations, change the path to the interpreter.

    .. versionchanged:: 1.1
        If *python_path* is None, the function does not change it.
    """
    if python_path is not None:
        # we change the path
        config = os.path.join(scite_path, "python.properties")
        with open(config, "r") as f:
            content = f.read()

        # we change the executable
        lines = content.split("\n")
        for i, line in enumerate(lines):
            if "command.go.*.py=" in line:
                lines[
                    i] = '    command.go.*.py={0} -u "$(FileNameExt)"'.format(python_path)
            elif "command.go.*.pyw=" in line:
                lines[
                    i] = '    command.go.*.pyw={0} -u "$(FileNameExt)"'.format(python_path)
        content = "\n".join(lines)
        with open(config, "w") as f:
            f.write(content)

    # we change the options
    config = os.path.join(scite_path, "SciTEGlobal.properties")
    with open(config, "r", encoding="utf8", errors="ignore") as f:
        content = f.read()
    content = content.replace("tabsize=8", "tabsize=4")
    content = content.replace("indent.size=8", "indent.size=4")
    content = content.replace("use.tabs=1", "use.tabs=0")
    content = content.replace("font:Verdana,", "font:Consolas,")
    with open(config, "w", encoding="utf8") as f:
        f.write(content)


def add_shortcut_to_desktop_for_scite(scite):
    """
    create a shortcut on your desktop

    @param      scite      scite location (SciTE.exe
    @return                filename
    """
    ver = suffix()
    return add_shortcut_to_desktop(scite, "SciTE." + ver, "SciTE." + ver)



# -*- coding: utf-8 -*-
"""
@file
@brief Defines different set of modules to install.


"""
from .packaged_config_0_minimal import minimal_set
from .packaged_config_1_small import small_set


_modules_set = [minimal_set, small_set]


def _function_set_name(f):
    """
    return the name of a function (not the module)

    @param      f       function
    @return             name

    .. versionadded:: 1.1
    """
    name = f.__name__
    return name.split(".")[-1]


def _build_set_correspondance(module_set):
    """
    builds a dictionary which returns a dictionary *{ name: function_set }*.

    @param      module_set      list of functions
    @return                     dictionary
    """
    res = {}
    for f in module_set:
        name = _function_set_name(f)
        res[name] = f
        name = name.replace("_set", "")
        res[name] = f
    return res


_module_set_name = _build_set_correspondance(_modules_set)


def get_package_set(name):
    """
    return a list of modules included in one the functions imported in this module

    @param      name        set name
    @return                 list of modules

    See :ref:`l-name-set-table` to get the list of available sets.
    """
    if name not in _module_set_name:
        keys = [_.replace("_set", "") for _ in _module_set_name.keys()]
        keys = list(sorted(set(keys)))
        raise ValueError("unable to find set for {0}\navailable sets:\n{1}".format(
            name, "\n".join(keys)))
    return _module_set_name[name]


def name_sets_dataframe(module_set=None):
    """
    returns a RST tables which described all the available sets

    @param      module_set      list of module sets or None to get all the sets described in this module
    @return                     list of dictionaries ``[ { "name": name1, "description": description } ]``

    the functions requires
    """
    if module_set is None:
        module_set = _modules_set

    def process_doc(doc):
        lines = [_.strip() for _ in doc.split("\n")]
        res = []
        for line in lines:
            if "@return" in line:
                break
            if not line.startswith(".. index::"):
                res.append(line)
        return (" ".join(res)).strip()

    res = [{"name": _function_set_name(f).replace("_set", ""),
            "description": process_doc(f.__doc__)} for f in module_set]
    so = [(_["name"], _) for _ in res]
    so.sort()
    return [_[1] for _ in so]


def classifiers2string(li):
    """
    Shortens the list of classifiers into a string.

    @param      li      list
    @return             string

    Example::

        ['Development Status :: 4 - Beta', 'Programming Language :: Python',
         'Programming Language :: Python :: 3', 'Programming Language :: Python :: 3',
         'Programming Language :: Python :: 3.9', 'License :: OSI Approved :: Apache Software License']
    """
    if li is None:
        return ""
    res = []
    for cl in sorted(set(li)):
        if cl == "Programming Language :: Python":
            s = "Python"
        else:
            s = cl.replace("Programming Language :: Python :: ", "")
            s = s.replace("License :: ", "")
            s = s.replace("OSI Approved :: ", "OSI Approved ")
            s = s.replace("Development Status :: ", "")
            s = s.replace("Operating System :: Microsoft :: ", "")
            s = s.replace("Operating System :: ", "")
            s = s.replace("Implementation :: ", "")
            s = s.replace("Programming Language :: ", "")
            s = s.replace("Environment :: ", "")
            s = s.replace("POSIX :: ", "")
            s = s.replace("BSD :: ", "")
        if s == "3 :: Only":
            s = "py3"
        elif s == "2 :: Only":
            s = "py2"
        elif s == "MacOS :: MacOS X":
            s = "MacOS/X"
        elif s == "POSIX :: Linux":
            s = "Linux"
        elif s == "POSIX :: BSD :: FreeBSD":
            s = "FreeBSD"
        elif s == "X11 Applications :: Qt":
            s = "Qt"
        elif s.startswith("Windows"):
            s = "Windows"
        if ("Framework" not in s and "Topic" not in s and
                "Intended Audience" not in s and "Natural Language" not in s):
            res.append(s)
    return ", ".join(res)



"""
@brief      test log(time=2s)
"""
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.packaged import small_set


class TestPackageList(unittest.TestCase):

    def test_package_list(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        filename = os.path.join(os.path.dirname(os.path.abspath(
            __file__)), "data", "status_installation_2016-05.txt")

        def clean(s):
            if "\\" in s or "/" in s:
                s = s.replace("\\", "/")[-1]
            return s.strip().lower().replace("_", "-")

        every = set(clean(_.name) for _ in small_set())

        collected = set()
        success = set()
        with open(filename, "r") as f:
            for line in f:
                line = line.strip("\n\r ")
                if "collected" in line:
                    pack = line.split(":")[-1].split(",")
                    for p in pack:
                        collected.add(clean(p))
                else:
                    end = line.split("installed")[-1]
                    spl = end.split()
                    for p in spl:
                        v = p.split('-')[0].strip()
                        if v:
                            success.add(clean(v))

        fLOG(len(collected), list(sorted(collected))[:4])
        fLOG(len(success), list(sorted(success))[:4])
        concat = collected.union(success)

        # not in the list
        fLOG('---------')
        not_list = concat - every
        for p in sorted(not_list):
            fLOG("not in", p)

        # not installed
        fLOG('---------')
        not_installed = every - concat
        for p in sorted(not_installed):
            fLOG("not installed", p)


if __name__ == "__main__":
    unittest.main()



"""
@file
@brief helpers around requirements
"""


def build_requirements(module_list):
    """
    builds a requirements list based on a list of @see cl ModuleInstall

    @param      module_list     list of @see cl ModuleInstall
    @return                     string
    """
    res = []
    for mod in module_list:
        if mod.version is not None:
            res.append("{0}=={1}".format(mod.name, mod.version))
        else:
            res.append(mod.name)
    return "\n".join(res)



"""
@brief      test log(time=20s)
"""

import sys
import os
import unittest
from pyquickhelper.loghelper import fLOG
from pymyinstall.installhelper.module_install import ModuleInstall


class TestDownloadPyQt (unittest.TestCase):

    def test_install_pyqt(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")
        fold = os.path.abspath(os.path.split(__file__)[0])
        temp = os.path.join(fold, "temp_download_pyqt")
        if not os.path.exists(temp):
            os.mkdir(temp)
        for _ in os.listdir(temp):
            if os.path.isfile(os.path.join(temp, _)):
                os.remove(os.path.join(temp, _))

        if sys.platform.startswith("win"):
            fLOG("install", "pyqt")
            m = ModuleInstall("PyQt5", "pip", mname="pyqt", fLOG=fLOG)
            exe = m.download(
                temp_folder=temp,
                file_save=os.path.join(
                    temp,
                    "out_page.html"), source="2")
            assert os.path.exists(exe)


if __name__ == "__main__":
    unittest.main()
